# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from thrift.transport import TSocket, TTransport

def regular_transport_factory(host, port, env, config_file):
    """
    Basic unencrypted Thrift transport factory function.
    Returns instantiated Thrift transport for use with cql.Connection.

    Params:
    * host .........: hostname of Cassandra node.
    * port .........: port number to connect to.
    * env ..........: environment variables (os.environ) - not used by this implementation.
    * config_file ..: path to cqlsh config file - not used by this implementation.
    """
    tsocket = TSocket.TSocket(host, port)
    return TTransport.TFramedTransport(tsocket)
